from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import numpy as np
from qiskit.primitives import BaseEstimator, EstimatorResult, PrimitiveJob

from .provider import BlueQubitBackendV2

if TYPE_CHECKING:
    from qiskit.circuit import QuantumCircuit
    from qiskit.quantum_info import SparsePauliOp

logger = logging.getLogger("bluequbit-python-sdk")


class BlueQubitEstimatorV1(BaseEstimator):
    """Evaluates expectation values for provided quantum circuit and observable combinations.

    Qiskit Estimator V1 interface to BlueQubit cloud computation. Compatible with
    ``qiskit_algorithms.VQE`` and other primitives that require the V1 estimator interface.

    .. deprecated::
        ``BaseEstimator`` (V1) is deprecated in Qiskit and will be removed in a future
        release. Prefer :class:`BlueQubitEstimatorV2` for new code.
    """

    def __init__(
        self,
        *,
        backend: BlueQubitBackendV2 | None = None,
        options: dict | None = None,
    ) -> None:
        """Initialize BlueQubitEstimatorV1.

        Args:
            backend: The BlueQubit backend to use. If ``None``, a new backend is
                created using environment variables or the provided options.
            options: Optional dict with keys ``"backend_options"`` (passed to
                ``BlueQubitBackendV2``) and ``"run_options"`` (passed to ``backend.run``).
        """
        super().__init__()
        self._custom_options = options if options is not None else {}
        if backend is None:
            backend = BlueQubitBackendV2(
                **self._custom_options.get("backend_options", {})
            )
        self._run_options = self._custom_options.get("run_options", {})
        self._backend = backend

    def _run(
        self,
        circuits: tuple[QuantumCircuit, ...],
        observables: tuple[SparsePauliOp, ...],
        parameter_values: tuple[tuple[float, ...], ...],
        **run_options: object,
    ) -> PrimitiveJob:
        job = PrimitiveJob(
            self._compute, circuits, observables, parameter_values, run_options
        )
        job._submit()  # noqa: SLF001
        return job

    def _compute(
        self,
        circuits: tuple[QuantumCircuit, ...],
        observables: tuple[SparsePauliOp, ...],
        parameter_values: tuple[tuple[float, ...], ...],
        run_options: dict,
    ) -> EstimatorResult:
        merged_run_options = {**self._run_options, **run_options}
        values = []
        metadata = []
        for circuit, observable, params in zip(
            circuits, observables, parameter_values, strict=False
        ):
            bound_circuit = circuit.assign_parameters(
                dict(zip(circuit.parameters, params, strict=True))
            )
            pauli_sums = list(zip(observable.paulis, observable.coeffs, strict=True))
            bq_job = self._backend.run(
                bound_circuit,
                pauli_sum=pauli_sums,  # type: ignore[arg-type]
                **merged_run_options,
            )
            result = bq_job.result()
            ev = np.real_if_close(result.expectation_value)
            values.append(ev)
            metadata.append({"circuit_metadata": circuit.metadata})
        return EstimatorResult(np.array(values, dtype=np.float64), metadata)



from math import ceil, log2

from qiskit import QuantumCircuit, QuantumRegister

from .draper_adder import DraperAdder


def multi_adder(m: int, n: int) -> QuantumCircuit:
    r"""A circuit that uses DraperAdder to perform in-place summation of :math:`m` qubit registers,
    each having :math:`n` qubits. The total sum will be in the last register, which should have extra
    qubits, initially at state :math:`|0\rangle`, to store the total maximum possible sum.

    :param m: The number of qubit registers, which must be at least 2
    :param n: The number of qubits in each register. It must be the same for all registers
    :raises ValueError: If ``m`` is lower than 2 or if ``n`` is lower than 1
    :return: circuit that performs summation of ``m`` registers

    :Example:

        Construct a circuit with m = 4 registers and n = 3 qubits for each. Set up the registers to be uniformly random
        numbers in their value range \(0-7\) and measure their sum.

        .. code-block:: python

            import bluequbit
            from bluequbit.library import multi_adder
            from qiskit import QuantumCircuit
            from math import ceil, log2

            m = 4
            n = 3
            num_sum_qubits = int(ceil(log2(m * (2**n - 1) + 0.5)))
            num_qubits = m * n + num_sum_qubits - n
            qc = QuantumCircuit(num_qubits, num_sum_qubits)
            qc.h(range(m * n))
            qc = qc.compose(multi_adder(m, n))
            qc.measure(range(num_qubits-num_sum_qubits, num_qubits), range(num_sum_qubits))

            bq = bluequbit.init("YOUR_TOKEN_HERE")
            result = bq.run(qc)
            print(result.get_counts())

    """
    if m < 2:
        raise ValueError("m must be at least 2")
    if n < 1:
        raise ValueError("n must be at least 1")

    max_reg_value = 2**n - 1
    max_sum = m * max_reg_value
    num_sum_qubits = ceil(log2(max_sum + 0.5))

    # create the registers
    q_registers = [QuantumRegister(n, name=f"q{i}") for i in range(m - 1)]
    last_qreg = QuantumRegister(num_sum_qubits, name=f"q{m - 1}")

    # build summation circuit
    sum_qc = QuantumCircuit(*q_registers, last_qreg)

    current_num_qubits = n
    current_max_value = 2**n
    current_max_sum = 2 * max_reg_value
    for qreg in q_registers:
        if current_max_sum < current_max_value:  # no need for a new qubit
            if (
                current_max_sum + max_reg_value < current_max_value
                and qreg != q_registers[-1]
            ):  # if next one is fixed
                sum_qc.append(
                    DraperAdder(
                        n, current_num_qubits, include_qft=False, include_iqft=False
                    ),
                    qreg[:] + last_qreg[:current_num_qubits],
                )
            else:  # if the next adder has kind="half"
                sum_qc.append(
                    DraperAdder(n, current_num_qubits, include_qft=False),
                    qreg[:] + last_qreg[:current_num_qubits],
                )
        else:  # add one qubit
            current_num_qubits += 1
            current_max_value *= 2
            if (
                current_max_sum + max_reg_value < current_max_value
                and qreg != q_registers[-1]
            ):  # if next one is fixed
                sum_qc.append(
                    DraperAdder(
                        n, current_num_qubits - 1, kind="half", include_iqft=False
                    ),
                    qreg[:] + last_qreg[:current_num_qubits],
                )
            else:  # if the next adder has kind="half"
                sum_qc.append(
                    DraperAdder(n, current_num_qubits - 1, kind="half"),
                    qreg[:] + last_qreg[:current_num_qubits],
                )
        current_max_sum += max_reg_value

    return sum_qc



import subprocess


def check_local_gpu_availability():
    try:
        subprocess.run(
            ["nvidia-smi"],  # noqa: S607
            check=True,
            capture_output=True,
            text=True,
            timeout=0.5,
        )
    except (
        subprocess.TimeoutExpired,
        subprocess.SubprocessError,
        FileNotFoundError,
    ):
        return False
    return True



from __future__ import annotations

import importlib.util
import logging
import os
import time
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any

from . import job_metadata_constants
from .circuit_serialization import encode_circuit_with_fallback, is_a_qiskit_circuit
from .computation_local import run_circuit_cirq, run_circuit_qiskit
from .exceptions import (
    BQJobNotCompleteError,
    BQSDKUsageError,
)
from .host_info import check_local_gpu_availability
from .job_result import JobResult
from .tools import format_pauli_sum

if TYPE_CHECKING:
    from collections.abc import Callable

logger = logging.getLogger("bluequbit-python-sdk")

_warning_no_deqart_internal_shown = False
_local_execution_mode_used = False


def run_circuits_local(
    circuits,
    device: str,
    *,
    job_name: str | None = None,
    shots: int | None = None,
    pauli_sum: Any | None = None,
    options: dict | None = None,
    tags: dict[str, Any] | None = None,
):
    global _warning_no_deqart_internal_shown  # noqa: PLW0603
    global _local_execution_mode_used  # noqa: PLW0603

    if not _local_execution_mode_used:
        os.environ["BLUEQUBIT_DEQART_INTERNAL_ANALYSIS_LOGGER_NAME"] = (
            "bluequbit-python-sdk"
        )
        _local_execution_mode_used = True

    # GPU availability check and fallback to CPU
    original_device = device
    if "gpu" in device:
        gpu_available = check_local_gpu_availability()
        if not gpu_available:
            device = job_metadata_constants.DEVICE_FALLBACK_MAPPING.get(device, device)
            logger.warning(
                f"NVIDIA GPU not detected locally. Falling back from '{original_device}' to '{device}' device."
            )

    # if changing device options, please change backend jobs.py new_single_job() code
    device_options = {}
    options = options or {}
    if "mps" in device:
        device_options["mps_bond_dimension"] = options.get(
            "mps_bond_dimension", job_metadata_constants.MPS_DEFAULT_BOND_DIMENSION
        )
        device_options["mps_truncation_threshold"] = options.get(
            "mps_truncation_threshold",
            job_metadata_constants.MPS_DEFAULT_TRUNCATION_THRESHOLD,
        )
    if "pauli-path" in device:
        pps_truncation_threshold = options.get("pauli_path_truncation_threshold")
        if pps_truncation_threshold is None:
            pps_truncation_threshold = 0.0
        device_options["pauli_path_truncation_threshold"] = pps_truncation_threshold
        pps_transpilation_level = options.get("pauli_path_circuit_transpilation_level")
        if pps_transpilation_level is None:
            pps_transpilation_level = (
                job_metadata_constants.PAULI_PATH_DEFAULT_TRANSPILATION_LEVEL
            )
        device_options["pauli_path_circuit_transpilation_level"] = (
            pps_transpilation_level
        )
        if "gpu" in device:
            pps_gpu_weight_cutoff = options.get("pauli_path_weight_cutoff")
            device_options["pauli_path_weight_cutoff"] = pps_gpu_weight_cutoff
    options = device_options

    created_on = time.time()
    if pauli_sum is not None:
        pauli_sum = format_pauli_sum(pauli_sum)

    def _format_result(r):
        def to_str_time(x):
            return (
                datetime.fromtimestamp(x, tz=timezone.utc)
                .astimezone()
                .strftime("%Y-%m-%d %H:%M:%S.%f")
            )

        r["created_on"] = to_str_time(created_on)
        r["run_start"] = to_str_time(r["run_start"])
        r["run_end"] = to_str_time(r["run_end"])
        r["has_statevector"] = "state_vector" in r
        r["run_status"] = "COMPLETED"
        r["job_name"] = job_name
        r["shots"] = shots
        r["pauli_sum"] = pauli_sum
        r["device_options"] = options
        r["tags"] = tags
        r["device"] = device
        if "mps" in r["device"]:
            r["run_results"] = {
                "mps_build_time": r["mps_build_time"],
            }

        res = JobResult(r)
        res._top_100k_results = r.get("counts")  # noqa: SLF001
        return res

    first_circuit = circuits[0] if isinstance(circuits, list) else circuits
    if importlib.util.find_spec("deqart_internal") is not None:
        # os.environ["BLUEQUBIT_PPS_USE_PARALLEL_COMPUTE"] = "1"
        from deqart_internal.analysis import estimate_and_validate
        from deqart_internal.computation import run_circuit

        logger.debug("Using local deqart_internal")

        def _run_and_format_result(circuit):
            encoded_circuit = encode_circuit_with_fallback(circuit)
            encoded_circuit["hybrid_code_snippet"] = None
            encoded_circuit["pauli_sum"] = pauli_sum
            encoded_circuit["shots"] = shots
            encoded_circuit["seed"] = None
            if options is not None:
                encoded_circuit["options"] = options
            (
                estimate_result,
                deqart_circuit,
                validation_result,
                validation_ok,
                _,
            ) = estimate_and_validate(
                encoded_circuit,
                False,
                device,
                execution_mode="local",
                decoded_circuit=circuit,
            )
            if not validation_ok:
                raise BQJobNotCompleteError(
                    "local",
                    "FAILED_VALIDATION",
                    validation_result,
                )
            r = run_circuit(deqart_circuit)
            jr = _format_result(r)
            jr.num_qubits_used = estimate_result["num_qubits_used"]
            jr.num_qubits = estimate_result["num_qubits"]
            jr.circuit = encoded_circuit
            jr.job_id = "local"
            return jr

        if isinstance(circuits, list):
            zero_fill_count = len(str(len(circuits) - 1))
            jrs = [_run_and_format_result(c) for c in circuits]
            for i, jr in enumerate(jrs):
                jr.batch_id = "local"
                if job_name is not None:
                    jr.job_name = f"{job_name}_{str(i).zfill(zero_fill_count)}"
            return jrs

        return _run_and_format_result(circuits)
    if not _warning_no_deqart_internal_shown:
        logger.warning("Using local execution mode without deqart_internal")
        _warning_no_deqart_internal_shown = True
    fn: Callable
    if is_a_qiskit_circuit(first_circuit):
        fn = run_circuit_qiskit
    elif str(type(first_circuit)) == "<class 'cirq.circuits.circuit.Circuit'>":
        fn = run_circuit_cirq
    else:
        error_msg = f"Circuit type not yet supported for {first_circuit}"
        raise BQSDKUsageError(error_msg) from None

    if isinstance(circuits, list):
        jrs = [_format_result(fn(c, shots=shots)) for c in circuits]
        for i, jr in enumerate(jrs):
            jr.batch_id = "localn"
            jr.job_id = "localn"
            if job_name is not None:
                jr.job_name = f"{job_name}_{str(i).zfill(len(str(len(circuits) - 1)))}"
            jr.num_qubits = len(circuits[i].qubits)
            jr.num_qubits_used = jr.num_qubits
        return jrs

    jr = _format_result(fn(circuits, shots=shots))
    jr.num_qubits = len(circuits.qubits)
    jr.num_qubits_used = jr.num_qubits
    jr.job_name = job_name
    jr.job_id = "localn"
    return jr



from .bell_pair import bell_pair
from .coin_toss import coin_toss
from .draper_adder import DraperAdder
from .ghz import ghz
from .multi_adder import multi_adder
from .qft import QFT
from .random_circuit import random_circuit
from .tfim_2d_grid import tfim_2d_grid

__all__ = [
    "QFT",
    "DraperAdder",
    "bell_pair",
    "coin_toss",
    "ghz",
    "multi_adder",
    "random_circuit",
    "tfim_2d_grid",
]
